#include <Peggy2.h>
#include "Characters.h"

// Frame split into 3 sub-frames (for 8 brightness levels)
Peggy2 frame4;
Peggy2 frame2;
Peggy2 frame1;

Peggy2 walls;
Peggy2 apples;

byte backingFrame; // frame that is being updated

// Define game area
const byte gameMinX = 0;
const byte gameMaxX = 24;
const byte gameMinY = 0;
const byte gameMaxY = 23;

// Define score area
const byte scoreMinX = 0;
const byte scoreMaxX = 24;
const byte scoreY = 24;

#define SNAKE 0
#define PONG 1
byte gameType = SNAKE;

// Define snake properties;
const byte snakeMaxLength = 128;
byte snakeX[snakeMaxLength];
byte snakeY[snakeMaxLength];
byte snakeHead;
byte snakeLength;
int snakeDirX;
int snakeDirY;
int points=0;
byte level=1;
byte levelMax=25; // Score to go to next level
byte speed=30;


void setup()
{
  frame4.HardwareInit();   // Call this once to init the hardware. 

  randomSeed(analogRead(5));
  // Setup buttons
  DDRB = 0xFE;    // PORTB pin 0 is button 0, set it up as an input
  PORTB |= 0x01;  // PORTB pin 0 pull up
  DDRC = 0xE0;    // PORTC pins 0 - 4 set up as inputs corresponding to buttons 1 - 5
  PORTC |= 0x1F;  // PORTC pins 0 - 4 pull ups
  
  frame4.Clear();
  frame2.Clear();
  frame1.Clear();

  display_intro();
  // Setup screen
  setup_level();
}

void loop()
{
  int newSnakeDirX=0;
  int newSnakeDirY=0;
  if (!(PINC & 0x02)) newSnakeDirX=-1;  // PORT C pin 1 is attached to left button
  if (!(PINC & 0x04)) newSnakeDirY=+1;  // PORT C pin 2 is attached to down button
  if (!(PINC & 0x08)) newSnakeDirY=-1;  // PORT C pin 3 is attached to up button
  if (!(PINC & 0x10)) newSnakeDirX=+1;  // PORT C pin 4 is attached to right button

  if (newSnakeDirX || newSnakeDirY) {
    if (!((snakeDirX == -newSnakeDirX) && (snakeDirY == -newSnakeDirY))) {
      snakeDirX=newSnakeDirX;
      snakeDirY=newSnakeDirY;
    }
  }

  move_snake();
  for (byte i=0;i<=speed;i++) {
    refresh_display();
  }
}

void setup_level()
{
  // Clear all frames
  frame4.Clear();
  frame2.Clear();
  frame1.Clear();
  apples.Clear();
  walls.Clear();

  // 5 initial apples
  for (byte i=0; i<5; i++) {
    add_apple();
  }

  add_walls();
  refresh_display();
}

void add_apple()
{
  byte appleX = random(gameMinX+1,gameMaxX);
  byte appleY = random(gameMinY+1,gameMaxY);
  if (frame1.GetPoint(appleX,appleY) || frame2.GetPoint(appleX,appleY) || frame4.GetPoint(appleX,appleY)) {
    add_apple();
    return;
  }
  apples.SetPoint(appleX,appleY);
  setPixel(appleX,appleY,4);
}

void add_walls()
{
  byte numLevels = 4;
  byte myLevel=((level-1) % numLevels) + 1;
  if (myLevel == 1) {
    // No walls
    setup_snake((gameMaxX-gameMinX)/2, (gameMaxY-gameMinY)/2);
  } 
  else if (myLevel == 2) {
    // Four outside walls
    add_wall(gameMinX,gameMinY,gameMaxX,gameMinY);
    add_wall(gameMaxX,gameMinY,gameMaxX,gameMaxY);
    add_wall(gameMinX,gameMinY,gameMinX,gameMaxY);
    add_wall(gameMinX,gameMaxY,gameMaxX,gameMaxY);
    setup_snake((gameMaxX-gameMinX)/2, (gameMaxY-gameMinY)/2);
  } 
  else if (myLevel == 3) {
    // Two walls splitting field into quarters
    add_wall((gameMaxX-gameMinX)/2, gameMinY, (gameMaxX-gameMinX)/2, gameMaxX);
    add_wall(gameMinX,(gameMaxY-gameMinY)/2, gameMaxX, (gameMaxY-gameMinY)/2);
    setup_snake(gameMinX+2, gameMinY+2);
  }
  else if (myLevel == 4) {
    // Four outside walls plus internal one
    add_wall(gameMinX,gameMinY,gameMaxX,gameMinY);
    add_wall(gameMaxX,gameMinY,gameMaxX,gameMaxY);
    add_wall(gameMinX,gameMinY,gameMinX,gameMaxY);
    add_wall(gameMinX,gameMaxY,gameMaxX,gameMaxY);
    add_wall((gameMaxX/3), ((gameMaxY-gameMinY)/2), (gameMaxX/3*2), ((gameMaxY-gameMinY)/2));
    setup_snake((gameMaxX-gameMinX)/2, (gameMaxY-gameMinY)/3); 
  }
}

void add_wall(byte x1, byte y1,byte x2, byte y2)
{
    setLine(x1,y1,x2,y2,7);
    walls.Line(x1,y1,x2,y2);
}

void setup_snake(byte x, byte y)
{
  for (byte i=0;i<snakeMaxLength;i++) {
    snakeX[i]=0xFF; // marker for unused spot
    snakeY[i]=0xFF; // marker for unused spot
  }
  snakeLength=1;
  snakeHead=snakeLength-1;
  snakeX[snakeHead] = x,
  snakeY[snakeHead] = y,
  setPixel(snakeX[snakeHead],snakeY[snakeHead],7);
  snakeDirX=random(3)-1;
  snakeDirY=random(3)-1;
}
void move_snake ()
{
  byte oldsnakeHead = snakeHead;
  byte oldsnakeTail=(snakeMaxLength+snakeHead-snakeLength) % snakeMaxLength;
  snakeHead=(snakeHead+1) % snakeMaxLength;

  byte snakeTail=(oldsnakeTail +1) % snakeMaxLength;

  snakeX[snakeHead] = (snakeX[oldsnakeHead] + snakeDirX) % (gameMaxX+1);
  if ((snakeX[snakeHead] < gameMinX) || (snakeX[snakeHead] > 50)) snakeX[snakeHead] = gameMaxX;

  snakeY[snakeHead] = (snakeY[oldsnakeHead] + snakeDirY) % (gameMaxY+1);
  if ((snakeY[snakeHead] < gameMinY) || (snakeY[snakeHead] > 50)) snakeY[snakeHead] = gameMaxY;

  // Check if hitting an apple
  if (apples.GetPoint(snakeX[snakeHead],snakeY[snakeHead])) {
    points++;
    speed = 32-((points % levelMax)/2)-(2*level);
    snakeLength=snakeLength+2;
    apples.ClearPoint(snakeX[snakeHead],snakeY[snakeHead]);
    add_apple();
    update_scorebar();
  }

  boolean hit = false;
  // Check if hiiting self
  for (byte i=0; i< snakeMaxLength; i++) {
    if (i != snakeHead) {
      if ((snakeX[i] == snakeX[snakeHead]) && (snakeY[i] == snakeY[snakeHead])) {
        hit=true;
      }
    }
  }
  // Check if hitting wall
  if (walls.GetPoint(snakeX[snakeHead],snakeY[snakeHead])) {
    hit=true;
  }

  if (hit) {
    display_score();
    points=0;
    level=1;
    setup_level();
    update_scorebar();
  } 
  else {
    // Set the head/tail values
    setPixel(snakeX[oldsnakeTail],snakeY[oldsnakeTail],0);
    snakeX[oldsnakeTail] = 0xFF;
    snakeY[oldsnakeTail] = 0xFF;
    setPixel(snakeX[snakeTail],snakeY[snakeTail],4);
    setPixel(snakeX[snakeHead],snakeY[snakeHead],7);
  }
}

void update_scorebar()
{
  byte mypoints = points % (levelMax);
  for (byte i=scoreMinX;i<=scoreMaxX;i++) {
    if ((i-scoreMinX) < points) {
      setPixel(i,scoreY,7);
    } 
    else {
      setPixel(i,scoreY,0);
    }
  }  
  if (points >= (level*levelMax)) {
    //next level
    level++;
    display_score();
    setup_level();
  }
}

void display_score()
{
  frame4.Clear();
  frame2.Clear();
  frame1.Clear();

  write_string("SCORE",0,0);
  char scoreChar[10];
  itoa(points, scoreChar, 10);
  write_string(scoreChar,0,7);

  write_string("LEVEL",1,18);
  char levelChar[10];
  itoa(level, levelChar, 10);
  write_string(levelChar,16,11);

  for (byte i=0; i< 100; i++) {
    refresh_display();
  }

  while((PINC & 0x1F) == 0x1F) {  // Wait for button 1-5
    refresh_display();
  }
}

void display_intro()
{
  byte selection=0xFF;
  write_string("SNAKE",1,2);
  
  while (PINC & 0x01) {  // Wait for button 1-5
        refresh_display();
    }
}

/*---------------------*
 *  Utility functions  *
 *---------------------*/

void setPixel(byte x, byte y, byte brightness)
{
  frame4.WritePoint(x,y,(brightness & 4));
  frame2.WritePoint(x,y,(brightness & 2));
  frame1.WritePoint(x,y,(brightness & 1));
}

void setLine(byte x1, byte y1, byte x2, byte y2, byte brightness)
{
  if (brightness & 4) frame4.Line(x1,y1,x2,y2);
  if (brightness & 2) frame2.Line(x1,y1,x2,y2);
  if (brightness & 1) frame1.Line(x1,y1,x2,y2);
}

void refresh_display ()
{
  frame4.RefreshAll(4);
  frame2.RefreshAll(2);
  frame1.RefreshAll(1);
}

void write_string(char string[], byte x, byte y)
{
  byte i=0;
  while (string[i] != '\0') {
    show_char(string[i], x+(5*i), y);
    i++;
  }
}

void show_char(char Character, byte x, byte y)
{
  byte charNum = 36;

  if ((Character >= '0') && (Character <= '9')) {
    charNum=Character-48;
  }
  if ((Character >= 'A') && (Character <= 'Z')) {
    charNum=Character-55;
  }

  unsigned long w;
  w = pgm_read_dword_near(chars + charNum);
  for (byte i=0; i<6; i++) {
    byte offset = 4*(5-i);
    byte b = w >> offset; // get four bits
    b=b & 0xF;
    if (b & B1000) setPixel(x,y+i,7);
    if (b & B0100) setPixel(x+1,y+i,7);
    if (b & B0010) setPixel(x+2,y+i,7);
    if (b & B0001) setPixel(x+3,y+i,7);
  }   
}

